#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>

#define STAT_PRIORITE_FAIBLE 1
#define STAT_PRIORITE_HAUTE  99

#define DYN_PRIORITE_FAIBLE 19
#define DYN_PRIORITE_HAUTE  -20

void affiche_infopriorite(void)
{
   printf("\tOrdonnancement SCHED_FIFO  : %d <= priorite <= %d\n",
         sched_get_priority_min (SCHED_FIFO),
         sched_get_priority_max (SCHED_FIFO));
   printf("\tOrdonnancement SCHED_RR    : %d <= priorite <= %d\n",
         sched_get_priority_min (SCHED_RR),
         sched_get_priority_max (SCHED_RR));
   printf("\tOrdonnancement SCHED_OTHER : %d <= priorite <= %d\n",
         sched_get_priority_min (SCHED_OTHER),
         sched_get_priority_max (SCHED_OTHER));
   printf("\n");
}

void affiche_ordonnancement(void) 
{
   struct sched_param schedparam;
   int ordonnancement;
   struct timespec quantum;

   ordonnancement = sched_getscheduler(0);
   if(ordonnancement < 0)
   {
      perror ("sched_getscheduler");
      exit (1);
   }
   if (sched_getparam(0, &schedparam) < 0) 
   {
      perror ("sched_getparam");
      exit (1);
   }
   switch (ordonnancement) 
   {
      case SCHED_FIFO: printf("\tOrdonnancement SCHED_FIFO : \n");
         printf("\t\tPriorité   = %d (sched_priority)\n", schedparam.sched_priority);
         break;
      case SCHED_OTHER: printf("\tOrdonnancement SCHED_OTHER :\n");
         printf("\t\tPriorité statique  = %d (sched_priority)\n", schedparam.sched_priority);
         printf("\t\tpriority dynamique = %d (getpriority)\n", getpriority(PRIO_PROCESS, 0));
         break;
      case SCHED_RR: printf("\tOrdonnancement SCHED_RR : \n");
         printf("\t\tPriorité   = %d (sched_priority)\n", schedparam.sched_priority);
         //calcul du rr
         sched_rr_get_interval(0, &quantum);
         printf("\t\tIntervalle = %u.%09u s (rr_interval)\n", (unsigned int)quantum.tv_sec, (unsigned int)quantum.tv_nsec);		
         break;
   }   
   printf("\n");
}

void modifie_ordonnancement(int policy, int priority, int sched_priority) 
{
   struct sched_param schedparam;   
   
   /* dynamique : priority est une valeur dans l'intervalle -20 (haute) à 19 (faible) */
   if (setpriority(PRIO_PROCESS, 0, priority) < 0) 
   {
      perror ("setpriority");
      exit (1);
   }
   /* statique : sched_priority est une valeur dans l'intervalle 1 (faible) à 99 (haute) */
   schedparam.sched_priority = sched_priority; 
   if (sched_setscheduler(0, policy, &schedparam) < 0) 
   {
      perror ("setscheduler");
      exit (1);
   }
}

int main(void) 
{
   printf("debut\n");

   printf("\t-> informations sur les priorités des ordonnancements connus\n");
   affiche_infopriorite();
   
   printf("\t-> informations l'ordonnancement de ce processus\n");
   affiche_ordonnancement();

   printf("\t-> modification de l'ordonnancement et des priorités de ce processus\n");
   //printf("\t   (il faut être root pour accéder à certaines de ces modifications)\n");
   //modifie_ordonnancement(SCHED_OTHER, DYN_PRIORITE_FAIBLE, 0);
   modifie_ordonnancement(SCHED_OTHER, DYN_PRIORITE_HAUTE, 0);   
   //modifie_ordonnancement(SCHED_RR, 0, STAT_PRIORITE_FAIBLE);
   //modifie_ordonnancement(SCHED_RR, 0, STAT_PRIORITE_HAUTE);    
   //modifie_ordonnancement(SCHED_FIFO, 0, STAT_PRIORITE_FAIBLE);
   //modifie_ordonnancement(SCHED_FIFO, 0, STAT_PRIORITE_HAUTE); 
   
   affiche_ordonnancement();

   printf("fin\n");
   exit(0);
}
