#define _GNU_SOURCE
#include <utmpx.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>

#define NB 500

#define STAT_PRIORITE 0
#define STAT_PRIORITE_FAIBLE 1
#define STAT_PRIORITE_HAUTE  99

#define DYN_PRIORITE 0
#define DYN_PRIORITE_FAIBLE 19
#define DYN_PRIORITE_HAUTE  -20

void syntaxe (char * nom)
{
   fprintf (stderr, "Syntaxe %s priorité\n", nom);
   exit (1);
}

void affiche_id(void) 
{
   fprintf (stdout, "PID = %u - PPID = %u - UID = %u - EUID = %u - CPU = %d\n", getpid(), getppid(), getuid(), geteuid(), sched_getcpu());
}

void affiche_ordonnancement(void) 
{
   struct sched_param schedparam;
   int ordonnancement;
   struct timespec quantum;

   affiche_id();
   ordonnancement = sched_getscheduler(0);
   if(ordonnancement < 0)
   {
      perror ("sched_getscheduler");
      exit (1);
   }
   if (sched_getparam(0, &schedparam) < 0) 
   {
      perror ("sched_getparam");
      exit (1);
   }
   switch (ordonnancement) 
   {
      case SCHED_FIFO: printf("\tOrdonnancement SCHED_FIFO : [%d]\n", getpid());
         printf("\t\tPriorité   = %d (sched_priority) [%d]\n", schedparam.sched_priority, getpid());
         break;
      case SCHED_OTHER: printf("\tOrdonnancement SCHED_OTHER : [%d]\n", getpid());
         printf("\t\tPriorité statique  = %d (sched_priority) [%d]\n", schedparam.sched_priority, getpid());
         printf("\t\tpriority dynamique = %d (getpriority) [%d]\n", getpriority(PRIO_PROCESS, 0), getpid());
         break;
      case SCHED_RR: printf("\tOrdonnancement SCHED_RR : [%d]\n", getpid());
         printf("\t\tPriorité   = %d (sched_priority) [%d]\n", schedparam.sched_priority, getpid());
         //calcul du rr
         sched_rr_get_interval(0, &quantum);
         printf("\t\tIntervalle = %u.%09u s (rr_interval) [%d]\n", (unsigned int)quantum.tv_sec, (unsigned int)quantum.tv_nsec, getpid());
         break;
   }   
   printf("\n");
}

void modifie_ordonnancement(int policy, int priority, int sched_priority) 
{
   struct sched_param schedparam;   
   
   /* dynamique : priority est une valeur dans l'intervalle -20 (haute) à 19 (faible) */
   if (setpriority(PRIO_PROCESS, 0, priority) < 0) 
   {
      perror ("setpriority");
      exit (1);
   }
   /* statique : sched_priority est une valeur dans l'intervalle 1 (faible) à 99 (haute) */
   schedparam.sched_priority = sched_priority; 
   if (sched_setscheduler(0, policy, &schedparam) < 0) 
   {
      perror ("setscheduler");
      exit (1);
   }
}

void modifie_cpu(int cpu)
{
   cpu_set_t mask; // un "ensemble de CPU"
   unsigned int len = sizeof(mask);
   pid_t pid = 0; /* le processus appelant */

   // lit le masque de CPU
   if (sched_getaffinity(pid, len, &mask) < 0) 
   {
      perror("sched_getaffinity");
      exit (1);
   }
   //printf("mask = %08lx\n", mask);
   
   CPU_ZERO(&mask); // vide le mask
   CPU_SET(cpu, &mask); // ajoute le CPU n°0
   //printf("mask = %08lx\n", mask);
   
   // écrit le nouveau masque de CPU
   if (sched_setaffinity(pid, len, &mask)) 
   {
      perror("sched_setaffinity");
      exit (1);
   }
}

/* les deux processus */
void etoile();
void diese();

int main (int argc, char * argv [])
{
   pid_t pidEtoile, pidDiese;
   int priorite;

   setbuf(stdout, NULL);
   // verification et recuperation de la priorite
   if ((argc < 2) || (sscanf (argv [1], "%d", & priorite) != 1))
      syntaxe (argv [0]);
 
   // rend le système mono-cpu
   modifie_cpu(0);
   
   // fixe la politique et la priorité demandee
   modifie_ordonnancement(SCHED_OTHER, priorite, 0); /* essayez 20 ! */
   affiche_ordonnancement();
   
   if ((pidEtoile = fork ()) == -1) { perror ("fork ()"); exit (1); }   
   if (pidEtoile == 0) etoile(); /* execute le processus fils etoile */
   else
   {      
      if ((pidDiese = fork ()) == -1) { perror ("fork ()"); exit (1); }      
      if (pidDiese  == 0) diese(); /* execute le processus fils diese */
      
      //fprintf (stdout, "Je suis le père et j'attends la fin de mes fils ...\n");
      wait (NULL);
      wait (NULL);
      fprintf (stdout, "Fin de mes fils ...\n");
   }
   
   return (0);
}
 
void etoile() 
{
   int i;
   char c1 = '*';

   fprintf (stdout, "Je suis le fils <etoile> : PID = %u\n", getpid());   
   affiche_ordonnancement();
   usleep(10000);
   
   for(i=1;i<=NB;i++)
   {
      write(1, &c1, 1);
      sched_yield();
   }
  
   sleep(1);
   fprintf (stdout, "\nJe suis le fils <etoile> et je me termine.\n\n");
   exit (0);  
}

void diese() 
{
   int i;
   char c1 = '#';
      
   fprintf (stdout, "Je suis le fils <diese> : PID = %u\n", getpid());   
   affiche_ordonnancement();
   usleep(10000);   
   
   for(i=1;i<=NB;i++)
   {
      write(1, &c1, 1);
      sched_yield();
   }
   
   sleep(1);
   fprintf (stdout, "\nJe suis le fils <diese> et je me termine.\n\n");
   exit (0);
}
